@@ -10,12 +10,12 @@
 ############################################################################
 
 import numpy as np
-from scipy import sparse
 
+from scipy import sparse
 import scqubits.utils.constants as constants
 import scqubits.utils.plotting as plot
 from scqubits.core.data_containers import WaveFunctionOnGrid
-from scqubits.core.discretization import GridSpec
+from scqubits.core.discretization import Grid1d, GridSpec
 from scqubits.core.qubit_base import QubitBaseClass
 from scqubits.utils.spectrum_utils import standardize_phases, order_eigensystem
 
@@ -78,9 +78,9 @@ class ZeroPi(QubitBaseClass):
 
         if EC is None and ECS is None:
             raise ValueError("Argument missing: must either provide EC or ECS")
-        elif EC and ECS:
+        if EC and ECS:
             raise ValueError("Argument error: can only provide either EC or ECS")
-        elif EC:
+        if EC:
             self.EC = EC
         else:
             self.EC = 1 / (1 / ECS - 1 / self.ECJ)
@@ -92,9 +92,10 @@ class ZeroPi(QubitBaseClass):
         self.grid = grid
         self.ncut = ncut
         self.truncated_dim = truncated_dim
-        self._sys_type = '0-Pi qubit without EL and EC disorder, no coupling to zeta mode'
+        self._sys_type = '0-pi'
         self._evec_dtype = np.complex_
-
+        self._default_var_range = (-np.pi / 2, 3 * np.pi / 2)
+        self._default_var_count = 100
 
     def _evals_calc(self, evals_count):
         hamiltonian_mat = self.hamiltonian()
@@ -104,17 +105,20 @@ class ZeroPi(QubitBaseClass):
     def _esys_calc(self, evals_count):
         hamiltonian_mat = self.hamiltonian()
         evals, evecs = sparse.linalg.eigsh(hamiltonian_mat, k=evals_count, return_eigenvectors=True, which='SA')
-
         # TODO consider normalization of zeropi wavefunctions
         # evecs /= np.sqrt(self.grid.grid_spacing())
-
         evals, evecs = order_eigensystem(evals, evecs)
         return evals, evecs
 
-    def ECS(self):
-        """ """
+    def get_ECS(self):
         return 1 / (1 / self.EC + 1 / self.ECJ)
 
+    def set_ECS(self, value):
+        raise ValueError("It's not possible to directly set ECS. Instead one can set EC or ECJ,\nor use "
+                         "set_EC_via_ECS() to update EC indirectly.")
+
+    ECS = property(get_ECS, set_ECS)
+
     def set_EC_via_ECS(self, ECS):
         """Helper function to set `EC` by providing `ECS`, keeping `ECJ` constant."""
         self.EC = 1 / (1 / ECS - 1 / self.ECJ)
@@ -124,12 +128,31 @@ class ZeroPi(QubitBaseClass):
         return self.grid.pt_count * (2 * self.ncut + 1)
 
     def potential(self, phi, theta):
-        return (-2.0 * self.EJ * np.cos(theta) * np.cos(
-            phi - 2.0 * np.pi * self.flux / 2.0) + self.EL * phi ** 2 + 2.0 * self.EJ +
-                self.EJ * self.dEJ * np.sin(theta) * np.sin(phi - 2.0 * np.pi * self.flux / 2.0))
+        """
+        Parameters
+        ----------
+        phi: float
+        theta: float
 
-    def sparse_kineticmat(self):
-        """ """
+        Returns
+        -------
+        float
+            value of the potential energy evaluated at phi, theta
+        """
+        return (-2.0 * self.EJ * np.cos(theta) * np.cos(phi - 2.0 * np.pi * self.flux / 2.0)
+                + self.EL * phi ** 2 + 2.0 * self.EJ
+                + self.EJ * self.dEJ * np.sin(theta) * np.sin(phi - 2.0 * np.pi * self.flux / 2.0))
+
+    def sparse_kinetic_mat(self):
+        """
+        Kinetic energy portion of the Hamiltonian.
+        TODO: update this method to use single-variable operator methods
+
+        Returns
+        -------
+        scipy.sparse.csc_matrix
+            matrix representing the kinetic energy operator
+        """
         pt_count = self.grid.pt_count
         dim_theta = 2 * self.ncut + 1
         identity_phi = sparse.identity(pt_count, format='csc', dtype=np.complex_)
@@ -137,19 +160,25 @@ class ZeroPi(QubitBaseClass):
 
         kinetic_matrix_phi = self.grid.second_derivative_matrix(prefactor=-2.0 * self.ECJ)
 
-        diag_elements = 2.0 * self.ECS() * np.square(np.arange(-self.ncut + self.ng, self.ncut + 1 + self.ng))
+        diag_elements = 2.0 * self.ECS * np.square(np.arange(-self.ncut + self.ng, self.ncut + 1 + self.ng))
         kinetic_matrix_theta = sparse.dia_matrix((diag_elements, [0]), shape=(dim_theta, dim_theta)).tocsc()
 
         kinetic_matrix = sparse.kron(kinetic_matrix_phi, identity_theta, format='csc') \
                          + sparse.kron(identity_phi, kinetic_matrix_theta, format='csc')
 
-        kinetic_matrix -= 2.0 * self.ECS() * self.dCJ * self.i_d_dphi_operator() * self.n_theta_operator()
-
+        kinetic_matrix -= 2.0 * self.ECS * self.dCJ * self.i_d_dphi_operator() * self.n_theta_operator()
         return kinetic_matrix
 
+    def sparse_potential_mat(self):
+        """
+        Potential energy portion of the Hamiltonian.
+        TODO: update this method to use single-variable operator methods
 
-    def sparse_potentialmat(self):
-        """Returns the potential energy matrix for the potential in sparse (`csc_matrix`) form."""
+        Returns
+        -------
+        scipy.sparse.csc_matrix
+            matrix representing the potential energy operator
+        """
         min_val = self.grid.min_val
         max_val = self.grid.max_val
         pt_count = self.grid.pt_count
@@ -162,26 +191,79 @@ class ZeroPi(QubitBaseClass):
         phi_sin_vals = np.sin(np.linspace(min_val, max_val, pt_count) - 2.0 * np.pi * self.flux / 2.0)
         phi_sin_potential = sparse.dia_matrix((phi_sin_vals, [0]), shape=(pt_count, pt_count)).tocsc()
 
-
-        theta_cos_potential = (-self.EJ * (sparse.dia_matrix(([1.0] * dim_theta, [-1]), shape=(dim_theta, dim_theta)) +
-                                           sparse.dia_matrix(([1.0] * dim_theta, [1]), shape=(dim_theta, dim_theta)))).tocsc()
+        theta_cos_potential = (-self.EJ
+                               * (sparse.dia_matrix(([1.0] * dim_theta, [-1]), shape=(dim_theta, dim_theta)) +
+                                  sparse.dia_matrix(([1.0] * dim_theta, [1]), shape=(dim_theta, dim_theta)))).tocsc()
         potential_mat = (sparse.kron(phi_cos_potential, theta_cos_potential, format='csc')
-                         + sparse.kron(phi_inductive_potential, self.identity_theta(), format='csc')
-                         + 2 * self.EJ * sparse.kron(self.identity_phi(), self.identity_theta(), format='csc'))
-        potential_mat += self.EJ * self.dEJ * sparse.kron(phi_sin_potential, self.identity_theta(), format='csc') \
+                         + sparse.kron(phi_inductive_potential, self._identity_theta(), format='csc')
+                         + 2 * self.EJ * sparse.kron(self._identity_phi(), self._identity_theta(), format='csc'))
+        potential_mat += self.EJ * self.dEJ * sparse.kron(phi_sin_potential, self._identity_theta(), format='csc')\
                          * self.sin_theta_operator()
-
         return potential_mat
 
     def hamiltonian(self):
-        """Returns Hamiltonian in basis obtained by discretizing phi and employing charge basis for theta"""
-        return self.sparse_kineticmat() + self.sparse_potentialmat()
+        """Calculates Hamiltonian in basis obtained by discretizing phi and employing charge basis for theta.
 
-    def identity_phi(self):
+        Returns
+        -------
+        scipy.sparse.csc_matrix
+            matrix representing the potential energy operator
+        """
+        return self.sparse_kinetic_mat() + self.sparse_potential_mat()
+
+    def sparse_d_potential_d_flux_mat(self):
+        r"""Calculates a of the potential energy w.r.t flux, at the current value of flux,
+        as stored in the object.
+
+        The flux is assumed to be given in the units of the ratio \Phi_{ext}/\Phi_0.
+        So if \frac{\partial U}{ \partial \Phi_{\rm ext}}, is needed, the expression returned
+        by this function, needs to be multiplied by 1/\Phi_0.
+
+        Returns
+        -------
+        scipy.sparse.csc_matrix
+            matrix representing the derivative of the potential energy
+        """
+        op_1 = sparse.kron(self._sin_phi_operator(x=- 2.0 * np.pi * self.flux / 2.0),
+                           self._cos_theta_operator(), format='csc')
+        op_2 = sparse.kron(self._cos_phi_operator(x=- 2.0 * np.pi * self.flux / 2.0),
+                           self._sin_theta_operator(), format='csc')
+        return - 2.0 * np.pi * self.EJ * op_1 - np.pi * self.EJ * self.dEJ * op_2
+
+    def d_hamiltonian_d_flux(self):
+        r"""Calculates a derivative of the Hamiltonian w.r.t flux, at the current value of flux,
+        as stored in the object.
+
+        The flux is assumed to be given in the units of the ratio \Phi_{ext}/\Phi_0.
+        So if \frac{\partial H}{ \partial \Phi_{\rm ext}}, is needed, the expression returned
+        by this function, needs to be multiplied by 1/\Phi_0.
+
+        Returns
+        -------
+        scipy.sparse.csc_matrix
+            matrix representing the derivative of the Hamiltonian
+        """
+        return self.sparse_d_potential_d_flux_mat()
+
+    def _identity_phi(self):
+        r"""
+        Identity operator acting only on the `\phi` Hilbert subspace.
+
+        Returns
+        -------
+            scipy.sparse.csc_matrix
+        """
         pt_count = self.grid.pt_count
         return sparse.identity(pt_count, format='csc')
 
-    def identity_theta(self):
+    def _identity_theta(self):
+        r"""
+        Identity operator acting only on the `\theta` Hilbert subspace.
+
+        Returns
+        -------
+            scipy.sparse.csc_matrix
+        """
         dim_theta = 2 * self.ncut + 1
         return sparse.identity(dim_theta, format='csc')
 
@@ -193,11 +275,12 @@ class ZeroPi(QubitBaseClass):
         -------
             scipy.sparse.csc_matrix
         """
-        return sparse.kron(self.grid.first_derivative_matrix(prefactor=1j), self.identity_theta(), format='csc')
+        return sparse.kron(self.grid.first_derivative_matrix(prefactor=1j), self._identity_theta(), format='csc')
 
-    def phi_operator(self):
+    def _phi_operator(self):
         r"""
-        Operator :math:`\varphi`.
+        Operator :math:`\varphi`, acting only on the `\varphi` Hilbert subspace.
+
 
         Returns
         -------
@@ -210,8 +293,17 @@ class ZeroPi(QubitBaseClass):
         phi_matrix = sparse.dia_matrix((pt_count, pt_count), dtype=np.complex_)
         diag_elements = np.linspace(min_val, max_val, pt_count)
         phi_matrix.setdiag(diag_elements)
+        return phi_matrix
 
-        return sparse.kron(phi_matrix, self.identity_theta(), format='csc')
+    def phi_operator(self):
+        r"""
+        Operator :math:`\varphi`.
+
+        Returns
+        -------
+            scipy.sparse.csc_matrix
+        """
+        return sparse.kron(self._phi_operator(), self._identity_theta(), format='csc')
 
     def n_theta_operator(self):
         r"""
@@ -224,8 +316,52 @@ class ZeroPi(QubitBaseClass):
         dim_theta = 2 * self.ncut + 1
         diag_elements = np.arange(-self.ncut, self.ncut + 1)
         n_theta_matrix = sparse.dia_matrix((diag_elements, [0]), shape=(dim_theta, dim_theta)).tocsc()
+        return sparse.kron(self._identity_phi(), n_theta_matrix, format='csc')
 
-        return sparse.kron(self.identity_phi(), n_theta_matrix, format='csc')
+    def _sin_phi_operator(self, x=0):
+        r"""
+        Operator :math:`\sin(\phi + x)`, acting only on the `\phi` Hilbert subspace.
+
+        Returns
+        -------
+            scipy.sparse.csc_matrix
+        """
+        min_val = self.grid.min_val
+        max_val = self.grid.max_val
+        pt_count = self.grid.pt_count
+
+        vals = np.sin(np.linspace(min_val, max_val, pt_count) + x)
+        sin_phi_matrix = sparse.dia_matrix((vals, [0]), shape=(pt_count, pt_count)).tocsc()
+        return sin_phi_matrix
+
+    def _cos_phi_operator(self, x=0):
+        r"""
+        Operator :math:`\cos(\phi + x)`, acting only on the `\phi` Hilbert subspace.
+
+        Returns
+        -------
+            scipy.sparse.csc_matrix
+        """
+        min_val = self.grid.min_val
+        max_val = self.grid.max_val
+        pt_count = self.grid.pt_count
+
+        vals = np.cos(np.linspace(min_val, max_val, pt_count) + x)
+        cos_phi_matrix = sparse.dia_matrix((vals, [0]), shape=(pt_count, pt_count)).tocsc()
+        return cos_phi_matrix
+
+    def _cos_theta_operator(self):
+        r"""
+        Operator :math:`\cos(\theta)`, acting only on the `\theta` Hilbert subspace.
+
+        Returns
+        -------
+            scipy.sparse.csc_matrix
+        """
+        dim_theta = 2 * self.ncut + 1
+        cos_theta_matrix = 0.5 * (sparse.dia_matrix(([1.0] * dim_theta, [-1]), shape=(dim_theta, dim_theta)) +
+                                  sparse.dia_matrix(([1.0] * dim_theta, [1]), shape=(dim_theta, dim_theta))).tocsc()
+        return cos_theta_matrix
 
     def cos_theta_operator(self):
         r"""
@@ -235,11 +371,21 @@ class ZeroPi(QubitBaseClass):
         -------
             scipy.sparse.csc_matrix
         """
-        dim_theta = 2 * self.ncut + 1
-        cos_theta_matrix = 0.5 * (sparse.dia_matrix(([1.0] * dim_theta, [-1]), shape=(dim_theta, dim_theta)) +
-                                  sparse.dia_matrix(([1.0] * dim_theta, [1]), shape=(dim_theta, dim_theta))).tocsc()
+        return sparse.kron(self._identity_phi(), self._cos_phi_operator(), format='csc')
 
-        return sparse.kron(self.identity_phi(), cos_theta_matrix, format='csc')
+    def _sin_theta_operator(self):
+        r"""
+        Operator :math:`\sin(\theta)`, acting only on the `\theta` Hilbert space.
+
+        Returns
+        -------
+            scipy.sparse.csc_matrix
+        """
+        dim_theta = 2 * self.ncut + 1
+        sin_theta_matrix = -0.5 * 1j * \
+                           (sparse.dia_matrix(([1.0] * dim_theta, [1]), shape=(dim_theta, dim_theta)) -
+                            sparse.dia_matrix(([1.0] * dim_theta, [-1]), shape=(dim_theta, dim_theta))).tocsc()
+        return sin_theta_matrix
 
     def sin_theta_operator(self):
         r"""
@@ -249,35 +395,32 @@ class ZeroPi(QubitBaseClass):
         -------
             scipy.sparse.csc_matrix
         """
-        dim_theta = 2 * self.ncut + 1
-        sin_theta_matrix = -0.5 * 1j * (sparse.dia_matrix(([1.0] * dim_theta, [1]), shape=(dim_theta, dim_theta)) -
-                                        sparse.dia_matrix(([1.0] * dim_theta, [-1]), shape=(dim_theta, dim_theta))).tocsc()
+        return sparse.kron(self._identity_phi(), self._sin_theta_operator(), format='csc')
 
-        return sparse.kron(self.identity_phi(), sin_theta_matrix, format='csc')
-
-    def plot_potential(self, theta_pts=100, contour_vals=None, aspect_ratio=None, filename=None):
+    def plot_potential(self, theta_range=None, theta_count=None, contour_vals=None, figsize=None, filename=None):
         """Draw contour plot of the potential energy.
 
         Parameters
         ----------
-        theta_pts: int, optional
-            (Default value = 100)
+        theta_range: tuple(float, float), optional
+            used for setting a custom plot range for theta
+        theta_count: int, optional
         contour_vals: list, optional
-            (Default value = None)
-        aspect_ratio: float, optional
-            (Default value = None)
+        figsize: tupel(float, float), optitional
         filename: str, optional
-            (Default value = None)
         """
+        theta_range, theta_count = self.try_defaults(theta_range, theta_count)
+
         min_val = self.grid.min_val
         max_val = self.grid.max_val
         pt_count = self.grid.pt_count
 
         x_vals = np.linspace(min_val, max_val, pt_count)
-        y_vals = np.linspace(-np.pi / 2, 3 * np.pi / 2, theta_pts)
-        return plot.contours(x_vals, y_vals, self.potential, contour_vals, aspect_ratio, filename)
+        y_vals = np.linspace(-np.pi / 2, 3 * np.pi / 2, theta_count)
+        return plot.contours(x_vals, y_vals, self.potential, contour_vals=contour_vals, figsize=figsize,
+                             filename=filename)
 
-    def wavefunction(self, esys=None, which=0, theta_pts=100):
+    def wavefunction(self, esys=None, which=0, theta_range=None, theta_count=None):
         """Returns a zero-pi wave function in `phi`, `theta` basis
 
         Parameters
@@ -285,9 +428,11 @@ class ZeroPi(QubitBaseClass):
         esys: ndarray, ndarray
             eigenvalues, eigenvectors
         which: int, optional
-             index of desired wave function (Default value = 0)
-        theta_pts: int, optional
-            number of points to be used in the 2pi interval
+             index of desired wave function (default value = 0)
+        theta_range: tuple(float, float), optional
+            used for setting a custom plot range for theta
+        theta_count: int, optional
+            number of points to be used in the interval for theta
 
         Returns
         -------
@@ -299,25 +444,26 @@ class ZeroPi(QubitBaseClass):
         else:
             _, evecs = esys
 
+        theta_range, theta_count = self.try_defaults(theta_range, theta_count)
+
         pt_count = self.grid.pt_count
         dim_theta = 2 * self.ncut + 1
         state_amplitudes = evecs[:, which].reshape(pt_count, dim_theta)
 
         # Calculate psi_{phi, theta} = sum_n state_amplitudes_{phi, n} A_{n, theta}
         # where a_{n, theta} = 1/sqrt(2 pi) e^{i n theta}
-        n_vec = np.arange(-self.ncut, self.ncut+1)
-        theta_vec = np.linspace(-np.pi / 2, 3 * np.pi / 2, theta_pts)
-        a_n_theta = np.exp(1j * np.outer(n_vec, theta_vec)) / (2 * np.pi)**0.5
+        n_vec = np.arange(-self.ncut, self.ncut + 1)
+        theta_vec = np.linspace(*theta_range, theta_count)
+        a_n_theta = np.exp(1j * np.outer(n_vec, theta_vec)) / (2 * np.pi) ** 0.5
         wavefunc_amplitudes = np.matmul(state_amplitudes, a_n_theta).T
         wavefunc_amplitudes = standardize_phases(wavefunc_amplitudes)
 
         grid2d = GridSpec(np.asarray([[self.grid.min_val, self.grid.max_val, pt_count],
-                                      [-np.pi / 2, 3 * np.pi / 2, theta_pts]]))
-
+                                      [*theta_range, theta_count]]))
         return WaveFunctionOnGrid(grid2d, wavefunc_amplitudes)
 
-    def plot_wavefunction(self, esys=None, which=0, theta_pts=100, mode='abs', zero_calibrate=False, figsize=(10, 5),
-                          aspect_ratio=3, fig_ax=None):
+    def plot_wavefunction(self, esys=None, which=0, theta_range=None, theta_count=None, mode='abs',
+                          zero_calibrate=True, figsize=(6, 2.75), fig_ax=None):
         """Plots 2d phase-basis wave function.
 
         Parameters
@@ -325,17 +471,17 @@ class ZeroPi(QubitBaseClass):
         esys: ndarray, ndarray
             eigenvalues, eigenvectors as obtained from `.eigensystem()`
         which: int, optional
-            index of wave function to be plotted (Default value = (0)
-        theta_pts: int, optional
-            number of points to be used in the 2pi interval
+            index of wave function to be plotted (default value = (0)
+        theta_range: tuple(float, float), optional
+            used for setting a custom plot range for theta
+        theta_count: int, optional
+            number of points to be used in the interval for theta
         mode: str, optional
-            choices as specified in `constants.MODE_FUNC_DICT` (Default value = 'abs_sqr')
+            choices as specified in `constants.MODE_FUNC_DICT` (default value = 'abs_sqr')
         zero_calibrate: bool, optional
             if True, colors are adjusted to use zero wavefunction amplitude as the neutral color in the palette
         figsize: (float, float), optional
             figure size specifications for matplotlib
-        aspect_ratio: float, optional
-            aspect ratio for matplotlib
         fig_ax: Figure, Axes, optional
             existing Figure, Axis if previous objects are to be appended
 
@@ -343,25 +489,43 @@ class ZeroPi(QubitBaseClass):
         -------
         Figure, Axes
         """
-        modefunction = constants.MODE_FUNC_DICT[mode]
-        wavefunc = self.wavefunction(esys, theta_pts=theta_pts, which=which)
-        wavefunc.amplitudes = modefunction(wavefunc.amplitudes)
-        return plot.wavefunction2d(wavefunc, figsize=figsize, aspect_ratio=aspect_ratio, zero_calibrate=zero_calibrate,
-                                   fig_ax=fig_ax)
+        theta_range, theta_count = self.try_defaults(theta_range, theta_count)
 
-    def filewrite_params_h5(self, h5file_root):
-        """Write current qubit parameters into a given h5 data file.
+        modefunction = constants.MODE_FUNC_DICT[mode]
+        wavefunc = self.wavefunction(esys, theta_count=theta_count, which=which)
+        wavefunc.amplitudes = modefunction(wavefunc.amplitudes)
+        return plot.wavefunction2d(wavefunc, figsize=figsize, zero_calibrate=zero_calibrate, fig_ax=fig_ax)
+
+    def set_params_from_dict(self, meta_dict):
+        """Set object parameters by given metadata dictionary
 
         Parameters
         ----------
-        h5file_root: open h5py file
+        meta_dict: dict
         """
-        super().filewrite_params_h5(h5file_root)
-        self.grid.filewrite_params_h5(h5file_root)
+        for param_name, param_value in meta_dict.items():
+            if isinstance(param_value, (int, float, np.number)):
+                if param_name in self.grid.__dict__.keys():
+                    setattr(self.grid, param_name, param_value)
+                else:
+                    setattr(self, param_name, param_value)
 
-    def set_params_from_h5(self, h5file_root):
-        super().set_params_from_h5(h5file_root)
-        grid_attrs = h5file_root.get('grid').attrs
-        self.grid.min_val = grid_attrs['min_val']
-        self.grid.max_val = grid_attrs['max_val']
-        self.grid.pt_count = grid_attrs['pt_count']
+    @classmethod
+    def create_from_dict(cls, meta_dict):
+        """Set object parameters by given metadata dictionary
+
+        Parameters
+        ----------
+        meta_dict: dict
+        """
+        filtered_dict = {}
+        grid_dict = {}
+        for param_name, param_value in meta_dict.items():
+            if isinstance(param_value, (int, float, np.number)):
+                if param_name in ['min_val', 'max_val', 'pt_count']:
+                    grid_dict[param_name] = param_value
+                else:
+                    filtered_dict[param_name] = param_value
+        grid = Grid1d(**grid_dict)
+        filtered_dict['grid'] = grid
+        return cls(**filtered_dict)
